<?php
#------------------------------------------------------------------------------- 
# 
# File: parser.php
# Author: Jason Wall and Jeremy Wall
# Definition: The parser package has a simple and single job to do, and that
#             is to take a block of text and turn it into an ordered array of
#             tokens containing either a text block or a tag/tag block.
#             Each token will contain both the block and a flag noting what
#             kind of token it is (text or tag).  The package then returns the
#             array.
#
#             Revised by Jeremy Wall to fix errors when parsing undefined Tag
#             Start definitions.
#-------------------------------------------------------------------------------

class Parser {

  # Package Methods
  function parse_text($textblock, $tagID='BK')
	{
  	# Function wide variables
  	 #$textblock contains the block of text to be parsed.
  	 # $tagID contains the ID of the tag it is optional
		 
  	$TagStartText = "<$tagID"; # Start tag prefix (it gets passed in)
  	$TagEndText = "</$tagID"; # End tag prefix (it also gets passed in)
  	$tokens = array();				#this variable will contain the tokenized text.
  	 
  	# loop through the string tokenizing as you go.  Since you are actually removing
  	#  the pieces of text as you go, when you are finished, the string should be empty.
  	while ($textblock != "") {
  		# it becomes necessary to define these here, as each iteration through
  		#   the loop will redefine them.
  		
  		$token = array(); # token hash
  		$tagname = "";
  		$block = "";
  		$attributes = array(); # attribute hash
  
  		# check to see if the first part of the string is not a tag, ie doesn't start with the Start Tag Prefix..
  		if (substr($textblock, 0, strlen($TagStartText)) != $TagStartText) {
  		
  			# find the begining of the tag.
  			$endpos = strpos($textblock, $TagStartText);
  			# populate the text block
  			if ($endpos === false) {
  				# if there are no more template tags, put the remaining text into the block 
  				$block = substr($textblock, 0);
  				#remove the text block from $textblock.
  				$textblock = "";
  			} else {
  				# if there are tags remaining, put the text that is in front of tag into the block
  				$block = substr($textblock, 0, $endpos);
  				#remove the text block from $textblock.
  				$textblock = substr($textblock, $endpos);
  			}			
  
  			# create and populate the token
  			$token = array (
  				'type' => "text", 
  				'tagname' => "text_block", 
  				'block' => $block,	
  			);
  
  			#put the token at the bottom of the stack.
  			array_push($tokens, $token); 
  		} 
  		# check to see if the first part of the string is a tag
  		elseif (substr($textblock, 0, strlen($TagStartText)) == $TagStartText) {
  			
  			# get the tag from the string.
  			$tag = substr($textblock,0,strpos($textblock, '>') +1);
  			
  			# This is where we determine if the tag is a block or a single tag.
  			#  We do this by looking for the forward slash, if it doesn't find one
  			#  then it means the tag is a block.
  			if (strpos($tag, '/>') === false) { 
  				
  				# Get the tag name.
  				$tagname = substr($tag, strlen($TagStartText), strpos($tag, '>')-strlen($TagStartText));
  				if (strpos($tagname, ' ') === true) {
  					$tagname = substr($tagname, 0, strpos($tagname, ' '));
  				}
  							
  				# Get the tag block.
  				# But first remove the opening tag from $textblock.
  				$textblock = substr($textblock, strpos($textblock, '>')+1);
  				
  			
  				# before going on, you must check for nested tags identical to the current tag.
  				$open_tag = $TagStartText.$tagname;
  				$close_tag = $TagEndText.$tagname;
  				$open_len = strlen($open_tag);
  				$close_len = strlen($close_tag);
  				$ignore = 0;
  				$found_end = false;
  				$idx_open_tag = strpos($textblock,$open_tag);
  				$idx_close_tag = strpos($textblock,$close_tag);
  				
  				while (!$found_end) {
    					
  				if ($idx_open_tag === false or $idx_open_tag > $idx_close_tag) {
  						# This is a possible end tag, check to see if it is what 
  						# we are looking for
  						if ($ignore > 0 ) {
  							# this is not the end tag you are looking for...
  							$ignore--;
  							$idx_close_tag = strpos($textblock, $close_tag, $idx_close_tag+$close_len);
  						
  						} else {
  	  						# This is the end tag you are looking for.
  							$found_end = true;
  						}
    					} else {
  						# this is a nested open tag, it needs to be checked 
  						# to see if it is self closing.
    						$nested_tag = substr($textblock,$idx_open_tag, strpos($textblock, '>', $idx_open_tag)-$idx_open_tag+1);
    						if (strpos($nested_tag, '/>') === false) { 
  							# this is a nested block tag
  							$ignore += 1;
    						}
  						# find the next open tag.
  						$idx_open_tag = strpos($textblock, $open_tag, $idx_open_tag+$open_len);
    					}
  				}				
  				
  				$block = substr($textblock, 0, $idx_close_tag);				
  
  				# and remove the block of text from $textblock.
  				$textblock = substr($textblock, $idx_close_tag);
  				# remove the closing tag from $textblock.
  				$textblock = substr($textblock, strpos($textblock, '>')+1);
  				
  				# get the attributes.
  				$tag = substr($tag,strlen($TagStartText));
  				$tag = substr($tag, 0, strpos($tag, '>'));
  
  				$attributes = $this->parse_attributes($tag);
  		
  				# populate the token			
  				$token = array(
  					'type' => "container",
  					'tagname' => $tagname,
  					'block' => $block,
  					'attributes' => $attributes,	
  				);
  				
  				
  			} else {
  				# this parses a single tag out.
  
  				# Get the tag name.
  				$tagname = substr($tag, strlen($TagStartText), strpos($tag, '>')-(strlen($TagStartText)+1));
  				if (strpos($tagname, ' ') !== false) {
  					$tagname = substr($tagname, 0, strpos($tagname, ' '));
  				}
  				
  				# remove the tag from $textblock.
  				$textblock = substr($textblock, strpos($textblock, '>')+1);
  			
  				# get the attributes.
  				$tag = substr($tag,strlen($TagStartText));
  				$tag = substr($tag, 0, strpos($tag, '/>'));
					//echo $tag;
  				$attributes = $this->parse_attributes($tag);
  		
  				# populate the token
  				$token = array(
  					'type' => "single", 
  					'tagname' => $tagname,
  					'block' => " ",
  					'attributes' => $attributes,
  				);
  			}
  			
  			# put the token on the bottom of the stack.
  			array_push($tokens, $token);
  		}
  	} 
  			
  		
  	return $tokens;
  }
  
  
  # this function knows how to parse the attributes out of a textblock 
  #   and returns them in a hash.
  function parse_attributes ($atag) {
  			

  	#variable declarations
  	$attributes = array();
  
  	#remove the tagname and closing carat.
		if (strpos($atag, ' ') != 0) {
    	$atag = substr($atag, strpos($atag, ' ')+1);
    	  
    	while (trim($atag) != "") {
    		$atag  = trim($atag);
    		$key   = trim(substr($atag, 0, strpos($atag, '=')));
    		$atag  = trim(substr($atag, strpos($atag, '=')+1));
    		$quote = substr($atag,0,1);
    		$atag  = substr($atag,1);
    		$value = substr($atag, 0, strpos($atag, $quote));
    		$atag  = substr($atag, strpos($atag, $quote) + 1 );		
    	
    		$attributes[$key] = $value;
    	}
		}  		
  	#return the hash.
  	return $attributes;
  }

}

/*
=head1 DESCRIPTION

The parser package has a simple and single job to do, and that is to take a block 
of text and turn it into an ordered array of tokens containing either a text block 
or a single tag or tag block. Each token will contain both the block and a flag 
noting what kind of token it is (text, tag_block or tag).  The package then returns 
the array.


=head1 METHODS

=item parse_text()

The parse_text() function takes the template text, a start tag prefix, and an end
tag prefix, parses the appropriate tags and returns a token tree in the form of an
array of hashes, each hash with a variable number of elements, depending on its type.

For Example:
 
%text_token = (
				type => "text", 
				block => $block,	
			);

%block_tag_token = (
				type => "block_tag", 
				tagname => $tagname,
				block => $block,
				attributes => \%attributes,
			);	

%tag_token = (
				type => "tag", 
				tagname => $tagname,
				attributes => \%attributes,
			);

The attributes value is a reference to an attributes hash in the form of:

%attributes = (
				"attribute_name" => "attribute_value",
			);

Further Notes: The token tree returned by parse_text is not iterative, thus the 
  tags inside a tag_block will not be processed, and you will need to call the 
	parse_text() function again when dealing with those tags.
	
	To build a complete tree, call the parse_text() function iteratively like so:
	
	sub parse {
		$template_text = shift;
		$start_tag_prefix = shift;
		$end_tag_prefix = shift;
		@token_tree;
	
    @tokens = Parser::parse_text($template_text,$start_tag_prefix,$end_tag_prefix);
	
    foreach $token (@tokens) {
        if ($token->{type} eq 'block_tag') {
            @sub_tokens;
            @sub_tokens = parse($token->{block},$start_tag_prefix,$end_tag_prefix);
            $token->{block} = \@sub_tokens;
        }
        push @token_tree, $token;
    }
    return @token_tree;
	}

=head1 AUTHOR

(c) 2004 Jason Wall, <jason@walljm.com>, www.walljm.com

=cut
*/
?>
